#!/usr/bin/env python

import os
import rgma
import sys

HIST_RETEN_TOKEN = 'HistoryRetentionPeriod'
TABLES_TOKEN = 'Tables'
TYPE_TOKEN = 'Type'
DBLOGICALNAME_TOKEN = 'DBLogicalName'
DBNAME_TOKEN = 'DBName'
DBPSWD_TOKEN = 'DBPassword'
DBUSER_TOKEN = 'DBUsername'

def help():
    script = 'flexible-archiver'
    print
    print "This 'swiss army knife' tool allows users to create and " + \
        "destroy any number of RGMA secondary producers. The ping option " + \
        'can then be used to keep any secondary producers alive on the ' + \
        'RGMA server.'
    print
    print 'Usage: ' + script + ' [--start --config <file> --output <file>] | ' + \
        '[--stop --input <file>] | [--ping --input <file>]'
    print
    print 'Options:'
    print '    --help,   -h'
    print '    --config, -c <config file used when creating an archiver, see --start>'
    print '    --ping,   -p <pings the archiver defined in id file, see --input>'
    print '    --start,  -s <creates an archiver defined by its config, see --config and --output>'
    print '    --stop,   -d <stops the archiver defined in the id file, see --input>'
    print '    --output, -o <output file where the created archiver id is stored>'
    print "    --input,  -i <file containing archiver id>"
    print
    print 'Examples:'
    print '    %s --start --config flexy.props --output archiver.id' % script
    print '    %s --ping --input archiver.id' % script
    print '    %s --stop --input archiver.id' % script
    print
    
def start(config):
    
    """ Kick starts the archiver, returning the resource 
        endpoint if creation was successfull. """
    
    interval = rgma.TimeInterval(75, rgma.Units.MINUTES)
    storage = None
    if config.has_key(DBLOGICALNAME_TOKEN):
        storage = rgma.Storage(rgma.StorageType.DATABASE, config[DBLOGICALNAME_TOKEN])
    elif config.has_key(DBPSWD_TOKEN) and config.has_key(DBNAME_TOKEN) \
        and config.has_key(DBUSER_TOKEN):
        storage = rgma.Storage(rgma.StorageType.DATABASE, config[DBNAME_TOKEN], 
            config[DBUSER_TOKEN], config[DBPSWD_TOKEN])
    else:
        storage = rgma.Storage(rgma.StorageType.DATABASE)
    props = rgma.ProducerProperties(storage, config[TYPE_TOKEN])
    endpoint = None
    try:
        sp = rgma.SecondaryProducer(interval, props)
        endpoint = sp.getEndpoint()
        for table in config[TABLES_TOKEN]:
            key = table + '_' + HIST_RETEN_TOKEN
            histRet = None
            if config.has_key(key):
                histRet = rgma.TimeInterval(int(config[key]), 
                    rgma.Units.MINUTES)
            else:
                histRet = rgma.TimeInterval(int(config[HIST_RETEN_TOKEN]), 
                    rgma.Units.MINUTES)
            sp.declareTable(table, '', histRet)
        print 'created archiver %s %s, credentials stored in output file' % \
            (endpoint.getUrl(), endpoint.getResourceId())
    except rgma.RGMAException, e:
        sys.stderr.write("Could not create archiver, error " + \
            "message: '%s'\n" % str(e))
        if (endpoint):
            try:
                sp.close()
            except rgma.RGMAException, e:
                pass
        sys.exit(1)
    return endpoint
    
def stop(endpoint):
    
    """ Closes the resource identified by its endpoint. Deletes
        the archiver id file. """
    
    try:
        print 'closing resource: %s %s' % (endpoint.getUrl(), \
            endpoint.getResourceId())
        resource = rgma.reconnect(rgma.SecondaryProducer, endpoint)
        resource.close()
    except Exception, e:
        sys.stderr.write("Error occured while connecting to resource, " +
            "error message: '%s'\n" % str(e))

def ping(endpoint):
    
    """ Sends a 'showSignOfLife' to the resource identified by its endpoint. 
        If the request fails, the script will exit with error code 1. """
        
    try:
        print 'pinging resource: %s %s' % (endpoint.getUrl(), \
            endpoint.getResourceId())
        resource = rgma.reconnect(rgma.SecondaryProducer, endpoint)
        resource.showSignOfLife()
    except Exception, e:
        sys.stderr.write("Error occured while connecting to resource, " +
            "error message: '%s'\n" % str(e))
        sys.exit(1)
    
def storeEndpoint(endpint, fname):
    
    """ Writes the endpoint to the file, stored as (url id)."""
    
    url = endpoint.getUrl()
    id = endpoint.getResourceId()
    try:
        f = open(fname, 'w')
        f.write('# Flexy id file - Do not touch!\n')
        f.write('%s %s\n' % (endpoint.getUrl(), endpoint.getResourceId()))
        f.close()
    except IOError, e:
        sys.stderr.write(str(e) + '\n')
        sys.exit(1)
        
def getEndpoint(fname):
    
    """ Returns the endpoint derived from the id file. Its assumed 
        that the file will contain one endpoint only, otherwise the
        script will exit with error code 1. """
    
    endpoints = []
    if os.path.isfile(fname):
        try:
            f = open(fname)
            lines = f.readlines()
            for line in lines:
                line = line.strip()
                if len(line) > 0 and not line.startswith('#'):
                    values = line.split()
                    resource = rgma.ResourceEndpoint(values[0], 
                        int(values[1]))
                    endpoints.append(resource)
        except IOError, e:
            sys.stderr.write('Error reading %s : %s\n' % (fname, str(e)))
            sys.exit(1)
    else:
        sys.stderr.write("Error: id file '%s' does not exist\n" % fname)
        sys.exit(1)
    if len(endpoints) == 0:
        sys.stderr.write("Error: id file '%s' contains no enpoint\n" % fname)
        sys.exit(1)
    if len(endpoints) > 1:
        sys.stderr.write("Error: found multiple endpoints in " + \
            "'%s', only 1 can be defined per file\n" % fname)
        sys.exit(1)
    return endpoints[0]
    
def parseConfigFile(fname):
    
    """ Parses the config file, returning any properties as a hashmap. """
    
    config={}
    if os.path.isfile(fname):
        try:
            f = open(fname)
            lines = f.readlines()
            f.close()
            currentKey = None
            for line in lines:
                line = line.strip()
                if line and not line.startswith('#'):
                    if line.rfind('=') == -1:
                        config[key] = config[key] + ' ' + line
                    else:
                        key, value = line.split('=')
                        config[key.strip()] = value.strip()
                        currentKey = key
        except IOError, e:
            sys.stderr.write('Error reading %s : %s\n' % (fname, str(e)))
            sys.exit(1)
    return config
    
def removeDuplicates(tableNames):
    
    """ Returns a list with all duplicate table names removed. """
    
    d = {}
    for name in tableNames:
        d[name] = None
    return d.keys()
    
def checkConfig(config, fname):
    
    """ Checks the user defined properties contain sensible 
        values. Terminates script if any errors are found. """
    
    error = False
    if config.has_key(HIST_RETEN_TOKEN) and len(config[HIST_RETEN_TOKEN]) == 0:
        sys.stderr.write("Error: '%s' is undefined\n" % HIST_RETEN_TOKEN)
        error = True
    if config.has_key(TABLES_TOKEN):
        if len(config[TABLES_TOKEN]) == 0:
            sys.stderr.write("Error: '%s' is undefined\n" % TABLES_TOKEN)
            error = True
        else:
            tables = removeDuplicates(config[TABLES_TOKEN].split())
            config[TABLES_TOKEN] = tables
            for table in tables:
                key = table + '_' + HIST_RETEN_TOKEN
                if config.has_key(key):
                    if len(config[key]) == 0:
                        sys.stderr.write("Error: '%s' must be defined\n" % key)
                        error = True
                elif not config.has_key(HIST_RETEN_TOKEN):
                    sys.stderr.write("Error: '%s' property is missing\n" % key)
                    error = True
    else:
        sys.stderr.write("Error: '%s' property is missing\n" % TABLES_TOKEN)
        error = True
    if config.has_key(TYPE_TOKEN):
        qtype = config[TYPE_TOKEN].upper()
        if qtype == 'LATEST':
            config[TYPE_TOKEN] = rgma.QueryType.LATEST
        elif qtype == 'HISTORY':
            config[TYPE_TOKEN] = rgma.QueryType.HISTORY
        else:
            sys.stderr.write("Error: '%s' must be either 'Latest' or 'History'\n" % \
                TYPE_TOKEN)
            error = True
    else:
        sys.stderr.write("Error: '%s' property is missing\n" % TYPE_TOKEN)
        error = True
    if error:
        sys.stderr.write("Check %s before retrying\n" % fname)
        sys.exit(1)
    
if __name__ == '__main__':
    
    START = 1
    STOP = 2
    PING = 3
    
    idfile = None
    outfile = None
    configfile = None
    opt = None
    previous = ''
    for arg in sys.argv[1:]:
        arg = arg.strip()
        if previous == '-i' or previous == '--input':
            idfile = arg
        if previous == '-c' or previous == '--config':
            configfile = arg
        if previous == '-o' or previous == '--output':
            outfile = arg
        elif arg == "-p" or arg == "--ping":
            opt = PING
        elif arg == "-d" or arg == "--stop":
            opt = STOP
        elif arg == "-s" or arg == "--start":
            opt = START
        elif arg == "-h" or arg == "--help":
            help()
            sys.exit(0)
        previous = arg
    if opt == STOP or opt == PING:
        if idfile == None:
            sys.stderr.write('Error: no id file defined\n')
            sys.exit(1)
        endpoint = getEndpoint(idfile)
        if opt == STOP:
            stop(endpoint)
            os.remove(idfile)
        else:
            ping(endpoint)
    elif opt == START:
        if configfile == None:
            sys.stderr.write('Error: no config file defined\n')
            sys.exit(1)
        if outfile == None:
            sys.stderr.write('Error: no output id file defined\n')
            sys.exit(1)
        if os.path.isfile(outfile):
            sys.stderr.write("Error, output id file '%s' already exists\n" % outfile)
            sys.exit(1)
            # Check if the file can be created - then delete it?
            
        config = parseConfigFile(configfile)
        checkConfig(config, configfile)
        endpoint = start(config)
        storeEndpoint(endpoint, outfile)
    else:
        help()
        sys.exit(1)
    sys.exit(0)

